#!/usr/bin/ruby

require 'loggability'
require 'pathname'
require 'redcloth'
require 'mail'

require 'strelka/app'
require 'strelka/mixins'

require 'strelka/cms'


class Strelka::CMS::Publisher < Strelka::App
	extend Loggability,
	       Configurability,
	       Strelka::MethodUtilities

	# Loggability API -- log to Strelka::CMS's logger
	log_to :strelka_cms

	# Configurability API -- use values from the 'cms' section of the config
	config_key :content_manager


	# Strelka App ID
	ID = 'content-manager'

	# The default directory to scan for pages
	DEFAULT_CONFIG = {
		pageroot: Pathname( __FILE__ ).dirname.parent.parent.parent + 'public'
	}

	# Pattern for untainting the page path
	PAGE_PATH_PATTERN = %r{
		\A
			(?<path>(?:
				[\w\-/]                       # Allow word characters, slashes, and hyphens
				|
				\.(?!\.)                      # or periods not followed by another period
			)*)
			(?:\.(?<suffix>html|page))?       # optional suffix
		\z
	}x


	##
	# The configured catalog root
	singleton_attr_accessor :pageroot

	##
	# The default content-type
	default_type 'text/html'


	# Strelka App setup
	plugin :templating
	layout 'layout.tmpl'
	templates \
		page:  'page.tmpl'

	# Handle errors using a templated view instead of the default text
	plugin :fancyerrors


	### Configurability API -- configure the CMS class once the config is loaded.
	def self::configure( config=nil )
		if config && config.member?( :pageroot )
			self.log.debug "Config is: %p" % [ config ]
			self.pageroot = Pathname( config.pageroot )
		else
			self.pageroot = DEFAULT_CONFIG[:pageroot]
		end
	end


	### Create a new instance of the CMS handler application.
	def initialize( * )
		super
		@catalog = Strelka::CMS::PageCatalog.new( self.class.pageroot )
	end


	##
	# The Strelka::CMS::PageCatalog for the page root.
	attr_reader :catalog


	#
	# Handler Routes
	#
	plugin :routing


	# /
	# -> /.page
	# -> /index.page

	# /foo/
	# -> /foo.page
	# -> /foo/index.page

	# /foo/project
	# -> /foo/project.page
	# -> /foo/project/index.page
	# -> /foo/project.html (literal)
	# -> /foo/project/index.html (literal)

	### GET /.*
	get do |request|
		path = self.get_page_path( request )

		# If there was a matching .page file, use that
		if page = self.get_page_for( path )
			self.log.debug "  found page: %p" % [ page ]
			return self.make_page_response( page )

		# Try a literal HTML page
		elsif htmlpage = self.get_html_path( path )
			return self.make_raw_response( request, htmlpage )

		else
			# Give up
			self.log.error "No pages matching %p" % [ path ]
			finish_with HTTP::NOT_FOUND
		end
	end


	### Return the page requested by the specified +request+ as a Pathname relative to the
	### application path.
	def get_page_path( request )
		unless path = request.app_path[ PAGE_PATH_PATTERN, :path ]
			self.log.error "Invalid app_path: %p" % [ request.app_path ]
			finish_with HTTP::NOT_FOUND
		end

		# Force the path to be relative and clean it up
		return Pathname( path.gsub(%r{\A\.?/+|/(?=/)|/+\z}, '') )
	end


	### Look for a static .html file in the catalog and return a Pathname for it if
	### one exists.
	def get_html_path( path )
		path = self.class.pageroot + path.sub_ext( '.html' )
		return path if path.exist?
		return nil
	end


	### Find a page that corresponds to the specified +path+ (a Pathname). Returns +nil+ if
	### no matching page was found.
	def get_page_for( path )
		self.log.debug "Page path is: %p" % [ path ]
		if path.to_s.empty?
			subcat = @catalog.matching_pattern( 'index.page' )
			return subcat.first
		else
			subcat = @catalog.relative_to( path.dirname ).
				matching_pattern( "#{path.basename(path.extname)}{.page,/index.page}" )
			return subcat.first
		end
	end


	### Package up the specified +page+ in the page template and return it.
	def make_page_response( page )
		tmpl = self.template( :page )
		tmpl.page = page
		return tmpl
	end


	### Package up the page at the specified +path+ in the response and return it.
	def make_raw_response( request, pagepath )
		response = request.response
		response.body = pagepath.open( 'r', encoding: 'utf-8' )
		return response
	end

end # class Strelka::CMS::Publisher

